from .TftEndpoint import TftEndpoint


class SummonerEndpoint(TftEndpoint):
    def __init__(self, url: str, **kwargs):
        nurl = "/summoner/v1/summoners" + url
        super().__init__(nurl, **kwargs)


class SummonerApiUrls:
    by_account = SummonerEndpoint("/by-account/{encrypted_account_id}")
    by_name = SummonerEndpoint("/by-name/{summoner_name}")
    by_puuid = SummonerEndpoint("/by-puuid/{puuid}")
    by_id = SummonerEndpoint("/{encrypted_summoner_id}")



from .. import BaseApi, NamedEndpoint
from .urls import ChampionMasteryApiV4Urls


class ChampionMasteryApiV4(NamedEndpoint):
    """
    This class wraps the Champion-Mastery-v4 Api calls provided by the Riot API.

    See https://developer.riotgames.com/api-methods/#champion-mastery-v4/ for more detailed
    information
    """

    def __init__(self, base_api: BaseApi):
        """
        Initialize a new ChampionMasteryApiV4 which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super().__init__(base_api, self.__class__.__name__)

    def by_summoner(self, region: str, encrypted_summoner_id: str):
        """
        Get all champion mastery entries.

        :param string region:                   the region to execute this request on
        :param string encrypted_summoner_id:    Summoner ID associated with the player

        :returns: List[ChampionMasteryDTO]: This object contains a list of Champion Mastery
                                            information for player and champion combination.
        """
        return self._request_endpoint(
            self.by_summoner.__name__,
            region,
            ChampionMasteryApiV4Urls.by_summoner,
            encrypted_summoner_id=encrypted_summoner_id,
        )

    def by_summoner_by_champion(
        self, region: str, encrypted_summoner_id: str, champion_id: int
    ):
        """
        Get a champion mastery by player ID and champion ID.

        :param string                           region: the region to execute this request on
        :param string encrypted_summoner_id:    Summoner ID associated with the player
        :param long champion_id:                Champion ID to retrieve Champion Mastery for

        :returns: ChampionMasteryDTO: This object contains single Champion Mastery information for
                                      player and champion combination.
        """
        return self._request_endpoint(
            self.by_summoner_by_champion.__name__,
            region,
            ChampionMasteryApiV4Urls.by_summoner_by_champion,
            encrypted_summoner_id=encrypted_summoner_id,
            champion_id=champion_id,
        )

    def scores_by_summoner(self, region: str, encrypted_summoner_id: str):
        """
        Get a player's total champion mastery score, which is the sum of individual champion
        mastery levels

        :param string region:                   the region to execute this request on
        :param string encrypted_summoner_id:    Summoner ID associated with the player

        :returns: int
        """
        return self._request_endpoint(
            self.scores_by_summoner.__name__,
            region,
            ChampionMasteryApiV4Urls.scores_by_summoner,
            encrypted_summoner_id=encrypted_summoner_id,
        )



import datetime
import logging
import threading

from collections import namedtuple
from typing import Iterable, Optional


log = logging.getLogger(__name__)

RawLimit = namedtuple("RawLimit", ["count", "limit", "time"])


class LimitCollection:
    def __init__(self):
        self._limits = {}
        self._limits_lock = threading.Lock()

    def wait_until(self) -> datetime.datetime:
        # we dont really want to update the limits as we process them
        # may be able to move the max() call outside the lock though
        with self._limits_lock:
            limits_waits = [limit.wait_until() for key, limit in self._limits.items()]
            return max(limits_waits) if limits_waits else None

    def update_limits(self, raw_limits: Iterable[RawLimit]):
        for raw_limit in raw_limits:
            if raw_limit.time not in self._limits:
                with self._limits_lock:
                    # check again in case it has already been added
                    if raw_limit.time not in self._limits:
                        self._limits[raw_limit.time] = Limit()
            self._limits[raw_limit.time].set_raw_limit(raw_limit)


class Limit:
    def __init__(self):
        self._start_time = None
        self._raw_limit = RawLimit(0, 0, 0)

        self._lock = threading.Lock()

    @property
    def start_time(self) -> Optional[datetime.datetime]:
        return self._start_time

    @property
    def duration(self) -> int:
        return self._raw_limit.time

    @property
    def count(self) -> int:
        return self._raw_limit.count

    @property
    def limit(self) -> int:
        return self._raw_limit.limit

    def set_raw_limit(self, raw_limit: RawLimit):
        enter_time = datetime.datetime.now()

        # try to ensure some thread safety
        with self._lock:
            reset_timer = False

            # if the time increment changed, we dont really know anything anymore
            # and should reset our timer
            if self._raw_limit.time != raw_limit.time:
                reset_timer = True
                if not (
                    self._raw_limit.time == 0
                    and self._raw_limit.limit == 0
                    and self._raw_limit.count == 0
                ):
                    log.warning(
                        "overwriting time limit, previously %s, now %s. This may cause rate limitting issues.",
                        self._raw_limit.time,
                        raw_limit.time,
                    )

            if self._raw_limit.limit != raw_limit.limit:
                log.info(
                    "rate limit changed from %s/%ss to %s/%ss",
                    self._raw_limit.limit,
                    self._raw_limit.time,
                    raw_limit.limit,
                    raw_limit.time,
                )

            # if the count is 1, that means we should reset our timers
            if raw_limit.count == 1:
                reset_timer = True

            if reset_timer:
                self._raw_limit = raw_limit
                self._start_time = enter_time
            else:
                # double check that we arent assigning a lower, non-1 value to
                # our count. This may be screwy if a bunch of requests are sent
                # as the rate limit is getting reset, but I dont think there
                # is a more elegant solution.
                if self._raw_limit.count > raw_limit.count:
                    raw_limit = RawLimit(
                        self._raw_limit.count, raw_limit.limit, raw_limit.time
                    )
                self._raw_limit = raw_limit

    def wait_until(self) -> datetime.datetime:
        if self.count >= self.limit:
            return self._start_time + datetime.timedelta(seconds=self.duration)
        return datetime.datetime(datetime.MINYEAR, 1, 1)



from .. import BaseApi, NamedEndpoint
from .urls import MatchApiUrls


class MatchApi(NamedEndpoint):
    """
    This class wraps the Val-Match-v1 Api calls provided by the Riot API.

    See https://developer.riotgames.com/apis#val-match-v1 for more detailed information
    """

    def __init__(self, base_api: BaseApi):
        """
        Initialize a new MatchApi which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super().__init__(base_api, self.__class__.__name__)

    def by_id(self, region: str, match_id: str):
        """
        Get match by id

        :returns: MatchDto
        """
        return self._request_endpoint(
            self.by_id.__name__, region, MatchApiUrls.by_id, match_id=match_id
        )

    def matchlist_by_puuid(self, region: str, puuid: str):
        """
        Get matchlist for games played by puuid

        :returns: MatchlistDto
        """
        return self._request_endpoint(
            self.matchlist_by_puuid.__name__,
            region,
            MatchApiUrls.matchlist_by_puuid,
            puuid=puuid,
        )

    def recent_matches(self, region: str, queue: str):
        """
        Get recent matches.

        Note: Returns a list of match ids that have completed in the last 10 minutes.
        NA/LATAM/BR share a match history deployment. As such, recent matches will return
        a combined list of matches from those three regions. Requests are load balanced so
        you may see some inconsistencies as matches are added/removed from the list.

        :returns: RecentMatchesDto
        """
        return self._request_endpoint(
            self.recent_matches.__name__,
            region,
            MatchApiUrls.recent_matches,
            queue=queue,
        )



from .RequestHandler import RequestHandler

from .DeprecationHandler import DeprecationHandler
from .DeserializerAdapter import DeserializerAdapter
from .DictionaryDeserializer import DictionaryDeserializer
from .RateLimiterAdapter import RateLimiterAdapter
from .TypeCorrectorHandler import TypeCorrectorHandler
from .ThrowOnErrorHandler import ApiError, ThrowOnErrorHandler



import re

from .. import BaseApi, Endpoint
from .urls import DataDragonUrls


class DataDragonApi:
    def __init__(self, base_api: BaseApi):
        self._base_api = base_api

    def champions(self, version: str, full: bool = False, locale=None):
        return self._request(
            DataDragonUrls.champions_full if full else DataDragonUrls.champions,
            version,
            locale,
        )

    def items(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.items, version, locale)

    def languages(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.languages, version, locale)

    def maps(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.maps, version, locale)

    def masteries(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.masteries, version, locale)

    def profile_icons(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.profile_icons, version, locale)

    def runes(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.runes, version, locale)

    def runes_reforged(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.runes_reforged, version, locale)

    def summoner_spells(self, version: str, locale: str = None):
        return self._request(DataDragonUrls.summoner_spells, version, locale)

    def versions_for_region(self, region: str):
        region = re.sub(r"\d", "", region)
        url, query = DataDragonUrls.versions(region=region)
        return self._base_api.raw_request_static(url, query)

    def _request(self, endpoint: Endpoint, version: str, locale: str):
        url, query = endpoint(version=version, locale=locale if locale else "en_US")
        return self._base_api.raw_request_static(url, query)



from .LeagueEndpoint import LeagueEndpoint


class SpecatorV4Endpoint(LeagueEndpoint):
    def __init__(self, url, **kwargs):
        nurl = f"/spectator/v4{url}"
        super().__init__(nurl, **kwargs)


class SpectatorApiV4Urls:
    by_summoner = SpecatorV4Endpoint(
        "/active-games/by-summoner/{encrypted_summoner_id}"
    )
    featured_games = SpecatorV4Endpoint("/featured-games")



from .. import BaseApi, NamedEndpoint
from .urls import ChampionApiV3Urls


class ChampionApiV3(NamedEndpoint):
    """
    This class wraps the Champion-v3 Api calls provided by the Riot API.

    See https://developer.riotgames.com/api-methods/#champion-v3 for more detailed information
    """

    def __init__(self, base_api: BaseApi):
        """
        Initialize a new ChampionApiV3 which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super().__init__(base_api, self.__class__.__name__)

    def rotations(self, region: str):
        """
        Returns champion rotations, including free-to-play and low-level free-to-play rotations.

        :returns: ChampionInfo
        """
        return self._request_endpoint(
            self.rotations.__name__, region, ChampionApiV3Urls.rotations
        )



from .TftEndpoint import TftEndpoint


class MatchEndpoint(TftEndpoint):
    def __init__(self, url: str, **kwargs):
        nurl = "/match/v1/matches" + url
        super().__init__(nurl, **kwargs)


class MatchApiUrls:
    by_puuid = MatchEndpoint("/by-puuid/{puuid}/ids", count=int)
    by_id = MatchEndpoint("/{match_id}")



from .TftEndpoint import TftEndpoint


class LeagueEndpoint(TftEndpoint):
    def __init__(self, url: str, **kwargs):
        nurl = "/league/v1" + url
        super().__init__(nurl, **kwargs)


class LeagueApiUrls:
    challenger = LeagueEndpoint("/challenger")
    by_summoner = LeagueEndpoint("/entries/by-summoner/{encrypted_summoner_id}")
    entries = LeagueEndpoint("/entries/{tier}/{division}", page=int)
    grandmaster = LeagueEndpoint("/grandmaster")
    by_id = LeagueEndpoint("/leagues/{league_id}")
    master = LeagueEndpoint("/master")
    rated_ladders = LeagueEndpoint("/rated-ladders/{queue}/top")



import datetime
import logging

from typing import Dict, Optional

from ...RateLimiter import RateLimiter

from . import ApplicationRateLimiter, MethodRateLimiter, OopsRateLimiter

log = logging.getLogger(__name__)


class BasicRateLimiter(RateLimiter):
    __application_rate_limiter = ApplicationRateLimiter()

    def __init__(self):
        super().__init__()

        self._limiters = (
            BasicRateLimiter.__application_rate_limiter,
            MethodRateLimiter(),
            OopsRateLimiter(),
        )

    def wait_until(
        self, region: str, endpoint_name: str, method_name: str,
    ) -> Optional[datetime.datetime]:
        wait_until = max(
            [
                (
                    limiter.wait_until(region, endpoint_name, method_name),
                    limiter.friendly_name,
                )
                for limiter in self._limiters
            ],
            key=lambda lim_pair: lim_pair[0]
            if lim_pair[0]
            else datetime.datetime(datetime.MINYEAR, 1, 1),
        )

        if wait_until[0] is not None and wait_until[0] > datetime.datetime.now():
            to_wait = wait_until[0] - datetime.datetime.now()

            log.debug(
                "waiting for %s seconds due to %s limit...",
                to_wait.total_seconds(),
                wait_until[1],
            )
            return wait_until[0]

    def record_response(
        self,
        region: str,
        endpoint_name: str,
        method_name: str,
        status: int,
        headers: Dict[str, str],
    ):
        for limiter in self._limiters:
            limiter.update_limiter(
                region, endpoint_name, method_name, status, headers,
            )



class UrlConfig:
    root_url = "https://{platform}.api.riotgames.com"
    lor_url = "https://{platform}.api.riotgames.com"
    riot_url = "https://{platform}.api.riotgames.com"
    tft_url = "https://{platform}.api.riotgames.com"
    val_url = "https://{platform}.api.riotgames.com"



from . import HeaderBasedLimiter


class ApplicationRateLimiter(HeaderBasedLimiter):
    def __init__(self):
        super().__init__("X-App-Rate-Limit", "X-App-Rate-Limit-Count", "Application")

    def _get_limit_scope(
        self, region: str, endpoint_name: str, method_name: str
    ) -> str:
        return f"{region}"



from .LorEndpoint import LorEndpoint


class MatchEndpoint(LorEndpoint):
    def __init__(self, url: str, **kwargs):
        super().__init__(f"/match/v1{url}", **kwargs)


class MatchApiUrls:
    by_puuid = MatchEndpoint("/matches/by-puuid/{puuid}/ids")
    by_id = MatchEndpoint("/matches/{match_id}")



from typing import Any

from requests import Response

from .RequestHandler import RequestHandler
from ..Deserializer import Deserializer


class DeserializerAdapter(RequestHandler):
    def __init__(self, deserializer: Deserializer):
        self._deserializer = deserializer

    def after_request(
        self,
        region: str,
        endpoint_name: str,
        method_name: str,
        url: str,
        response: Response,
    ) -> Any:
        return self._deserializer.deserialize(endpoint_name, method_name, response.text)

    def after_static_request(self, url: str, response: Response) -> Any:
        # TODO: add params to get endpoint_name and method_name
        return self._deserializer.deserialize(None, None, response.text)



from .LeagueEndpoint import LeagueEndpoint


class ClashV1Endpoint(LeagueEndpoint):
    def __init__(self, url, **kwargs):
        nurl = f"/clash/v1{url}"
        super().__init__(nurl, **kwargs)


class ClashApiV1Urls:
    by_summoner = ClashV1Endpoint("/players/by-summoner/{summoner_id}")
    by_team = ClashV1Endpoint("/teams/{team_id}")
    tournaments = ClashV1Endpoint("/tournaments")
    tournament_by_team = ClashV1Endpoint("/tournaments/by-team/{team_id}")
    by_tournament = ClashV1Endpoint("/tournaments/{tournament_id}")



from .RiotEndpoint import RiotEndpoint


class AccountEndpoint(RiotEndpoint):
    def __init__(self, url: str, **kwargs):
        super().__init__(f"/account/v1{url}", **kwargs)


class AccountApiUrls:
    by_puuid = AccountEndpoint("/accounts/by-puuid/{puuid}")
    by_riot_id = AccountEndpoint("/accounts/by-riot-id/{game_name}/{tag_line}")
    active_shard = AccountEndpoint("/active-shards/by-game/{game}/by-puuid/{puuid}")



import requests


class BaseApi:
    def __init__(self, api_key, request_handlers=None, timeout=None):
        self._api_key = api_key
        self._request_handlers = request_handlers
        self._timeout = timeout

    @property
    def api_key(self):
        return self._api_key

    def raw_request(
        self,
        endpoint_name: str,
        method_name: str,
        region: str,
        url: str,
        query_params: dict,
    ):
        query_params = {k: v for k, v in query_params.items() if v is not None}

        response = None
        early_ret_idx = None

        if self._request_handlers is not None:
            for idx, handler in enumerate(self._request_handlers, start=1):
                response = handler.preview_request(
                    region, endpoint_name, method_name, url, query_params
                )
                early_ret_idx = idx
                if response is not None:
                    break

        if response is None:
            extra = {}
            if self._timeout is not None:
                extra["timeout"] = self._timeout

            response = requests.get(
                url,
                params=query_params,
                headers={"X-Riot-Token": self.api_key},
                **extra
            )

        if self._request_handlers is not None:
            for handler in self._request_handlers[early_ret_idx:None:-1]:
                mod = handler.after_request(
                    region, endpoint_name, method_name, url, response
                )
                if mod is not None:
                    response = mod

        return response

    def raw_request_static(self, url, query_params):
        query_params = {k: v for k, v in query_params.items() if v is not None}

        response = None
        early_ret_idx = None

        if self._request_handlers is not None:
            for idx, handler in enumerate(self._request_handlers, start=1):
                response = handler.preview_static_request(url, query_params)
                early_ret_idx = idx
                if response is not None:
                    break

        if response is None:
            extra = {}
            if self._timeout is not None:
                extra["timeout"] = self._timeout

            response = requests.get(url, params=query_params, **extra)

        if self._request_handlers is not None:
            for handler in self._request_handlers[early_ret_idx:None:-1]:
                mod = handler.after_static_request(url, response)
                if mod is not None:
                    response = mod

        return response



from .. import BaseApi, NamedEndpoint
from .urls import LolStatusApiV3Urls


class LolStatusApiV3(NamedEndpoint):
    """
    This class wraps the LoL-Status-v3 Api calls provided by the Riot API.

    See https://developer.riotgames.com/api-methods/#lol-status-v3 for more detailed information
    """

    def __init__(self, base_api: BaseApi):
        """
        Initialize a new LolStatusApiV3 which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super().__init__(base_api, LolStatusApiV3.__name__)

    def shard_data(self, region: str):
        """
        Get League of Legends status for the given shard.

        Requests to this API are not counted against the application Rate Limits.

        :param string region: the region to execute this request on

        :returns: ShardStatus
        """
        return self._request_endpoint(
            self.shard_data.__name__, region, LolStatusApiV3Urls.shard_data
        )



from .LeagueEndpoint import LeagueEndpoint


class ChampionMasteryV4Endpoint(LeagueEndpoint):
    def __init__(self, url, **kwargs):
        nurl = f"/champion-mastery/v4{url}"
        super().__init__(nurl, **kwargs)


class ChampionMasteryApiV4Urls:
    by_summoner = ChampionMasteryV4Endpoint(
        "/champion-masteries/by-summoner/{encrypted_summoner_id}"
    )
    by_summoner_by_champion = ChampionMasteryV4Endpoint(
        "/champion-masteries/by-summoner/{encrypted_summoner_id}/by-champion/{champion_id}"
    )
    scores_by_summoner = ChampionMasteryV4Endpoint(
        "/scores/by-summoner/{encrypted_summoner_id}"
    )



from .__version__ import __author__, __title__, __version__

from .exceptions import ApiError, TimeoutError
from .Deserializer import Deserializer
from .RateLimiter import RateLimiter
from .LolWatcher import LolWatcher
from .LorWatcher import LorWatcher
from .TftWatcher import TftWatcher
from .ValWatcher import ValWatcher

from .riotwatcher import RiotWatcher

__all__ = [
    "Deserializer",
    "RateLimiter",
    "LolWatcher",
    "LorWatcher",
    "TftWatcher",
    "RiotWatcher",
    "ValWatcher",
    "Handlers",
    "ApiError",
    "TimeoutError",
]



from .. import BaseApi, NamedEndpoint
from .urls import MatchApiUrls


class MatchApi(NamedEndpoint):
    """
    This class wraps the TFT-Match-v1 Api calls provided by the Riot API.

    See https://developer.riotgames.com/apis#tft-match-v1 for more detailed information
    """

    def __init__(self, base_api: BaseApi):
        """
        Initializes a new MatchApi which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super().__init__(base_api, self.__class__.__name__)

    def by_puuid(self, region: str, puuid: str, count: int = 20):
        """
        Get a list of match ids by PUUID.

        :returns: List[string]
        """
        return self._request_endpoint(
            self.by_puuid.__name__,
            region,
            MatchApiUrls.by_puuid,
            puuid=puuid,
            count=count,
        )

    def by_id(self, region: str, match_id: str):
        """
        Get a match by match id.

        :returns: MatchDto
        """
        return self._request_endpoint(
            self.by_id.__name__, region, MatchApiUrls.by_id, match_id=match_id
        )



from .LeagueEndpoint import LeagueEndpoint


class MatchV4Endpoint(LeagueEndpoint):
    def __init__(self, url, **kwargs):
        nurl = f"/match/v4{url}"
        super().__init__(nurl, **kwargs)


class MatchApiV4Urls:
    by_id = MatchV4Endpoint("/matches/{match_id}")
    matchlist_by_account = MatchV4Endpoint(
        "/matchlists/by-account/{encrypted_account_id}",
        queue=(int,),
        beginTime=int,
        endTime=int,
        beginIndex=int,
        endIndex=int,
        season=(int,),
        champion=(int,),
    )
    timeline_by_match = MatchV4Endpoint("/timelines/by-match/{match_id}")



from .. import BaseApi, NamedEndpoint
from .urls import ThirdPartyCodeApiV4Urls


class ThirdPartyCodeApiV4(NamedEndpoint):
    """
    This class wraps the ThirdPartyCode-v4 Api calls provided by the Riot API.

    See https://developer.riotgames.com/api-methods/#third-party-code-v4 for more detailed
    information
    """

    def __init__(self, base_api: BaseApi):
        """
        Initialize a new ThirdPartyCodeApiV4 which uses the provided base_api

        :param BaseApi base_api: the root API object to use for making all requests.
        """
        super().__init__(base_api, self.__class__.__name__)

    def by_summoner(self, region: str, encrypted_summoner_id: str):
        """
        FOR KR SUMMONERS, A 404 WILL ALWAYS BE RETURNED.

        Valid codes must be no longer than 256 characters and only use
        valid characters: 0-9, a-z, A-Z, and -

        :param string region:                   the region to execute this request on
        :param string encrypted_summoner_id:    Summoner ID

        :returns: string
        """
        return self._request_endpoint(
            self.by_summoner.__name__,
            region,
            ThirdPartyCodeApiV4Urls.by_summoner,
            encrypted_summoner_id=encrypted_summoner_id,
        )
